/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

/*
If you want to use the SyncGenie as a standalone trigger-box, your best bet is
to upload `FourJacksMenuDriven.ino` and read the accompanying manual. If you
want to program the SyncGenie yourself, and take advantage of the event-detection
algorithms in the `SyncGenie` library, then keep reading here.

This sketch demonstrates some of the key classes of the `SyncGenie` library:

- `Trigger` (for oscillatory analog signals like audio),
- `TriggerDC` (for steady analog signals, e.g. from a light sensor),
- `Switch` (for normally-open switches and push-buttons),
- `Gauge` (for potentiometers and other analog readings that we do not want to
  threshold).

It builds a trigger system with four inputs/outputs:

- First output: a light sensor (photo-transistor) on TRS1 (`TriggerDC` class)
- Second output: an audio trigger on TRS2 (`Trigger` class)
- Third output: a TRS push-button ("nurse call" cord) on TRS3 (`Switch` class)
- Fourth output: the SyncGenie's built-in top-side push-button (`Switch` class)

To demonstrate the `Gauge` class, the SyncGenie's built-in potentiometers are
used to adjust the sensitivity of the light and sound detectors.

We are assuming the light detector is a photo-transistor that needs voltage to
be supplied on one contact (we use the ring of TRS1) while you read the analog
signal from another (we use the tip of TRS1). If you're using a light-dependent
RESISTOR (LDR) instead of a photo-transistor, you may have to tweak the code to
use much higher threshold values but you may still be able to use the same
approach (set one contact `HIGH` and do analog reads from the other, thresholded
through a `TriggerDC` instance). An additional possibility with an LDR is to set
one contact `LOW`, put the other one in `INPUT_PULLUP` mode and do digital reads
from it---a `Switch` instance would do this if you set it up as
`Switch ldr(genie.TRS1_TIP, LOW, genie.TRS1_RING);`

An almost-equivalent configuration via the `FourJacksMenuDriven.ino` interactive
menu would be:

- Configure TRS1 -> TRS1 Mode -> DC TRIGGER          
- Configure TRS1 -> TRS1 Circuit -> RING HI READ TIP  (photo-transistor or LDR)

or

- Configure TRS1 -> TRS1 Mode -> SWITCH             
- Configure TRS1 -> TRS1 Circuit -> RING LO READ TIP  (LDR only)

and then

- Configure TRS2 -> TRS2 Mode -> AC TRIGGER
- Configure TRS2 -> TRS2 Circuit -> RING LO READ TIP  (disables processing on ring)

- Configure TRS3 -> TRS3 Mode -> SWITCH
- Configure TRS3 -> TRS3 Circuit -> TIP HI READ RING

- Configure TRS4 -> TRS4 Mode -> NO PROCESSING        (disables both tip and ring)

- Sample Rate -> 10000 Hz
- Stored Settings -> SAVE NOW

(The differences are that the menu-driven sketch has no support for the top-side
push-button, that the thresholds would be adjusted via the menu rather than the
potentiometers, and that the output pins are somewhat different---although with
a bit more twiddling, you can reconfigure the output pins too.)

*/

#include <SyncGenie.h>
SyncGenie genie;  // this instance represents the SyncGenie device as a whole

const int lightSignalPin       = genie.TRS1_TIP;  // Let's read a light sensor (phototransistor) on the tip of TRS1.
const int lightSensorSupplyPin = genie.TRS1_RING; // We'll need to supply voltage to the photo-transistor, so let's do that on the ring of TRS1. 
#include <Triggering.h>
TriggerDC light( lightSignalPin, genie.D0_OUT );  
Trigger   sound( genie.TRS2_TIP, genie.D1_OUT );  // Let's also read an audio signal from the tip ("left" channel) of TRS2.
// NB: `Trigger` and `TriggerDC` instances default to a sampling rate of 10kHz which should be fine (depending on how much
// other processing you do in `loop()`) as long as you are using 5 or fewer of them. (On a Teensy 4.1 at standard processor
// speed, we find that sampling and processing takes 16-17 microseconds per instance per sample.)

#include <Switch.h>
Switch response(     // Let's connect a "nurse call" type TRS push-button to TRS3, for subject responses.
	genie.TRS3_RING, // We'll read from the ring contact,
	HIGH,            // and expect a HIGH signal when the button is pressed (it typically connects tip to ring through a 1kOhm resistor)
	genie.TRS3_TIP   // so that means we're using the tip as reference: the Switch constructor will therefore set it to output HIGH.
); // (you could equally well specify `LOW` instead of `HIGH`: then the Switch would automatically put the input pin into `INPUT_PULLUP` mode).
const int responseOutputPin = genie.D2_OUT;


// The `sound.threshold` and `light.threshold` parameters can be set any number of ways.
// For this example, let's use the SyncGenie's potentiometers to allow the user to adjust them directly by hand.
// Note that potentiometer readings will fluctuate randomly a little, so may not be ideal for this purpose in all
// applications; but the `Gauge` class (defined in Switch.h) smooths these changes somewhat, and they should have
// little impact here:
Gauge lightAdjustment( genie.POT1 );
Gauge soundAdjustment( genie.POT2 );


void setup()
{
	genie.begin( "Demo sketch:\nVariousDetectors" ); // this initializes some pin modes, starts the LCD screen, starts the Serial connection, and flickers D9_LED (same as LED_BUILTIN) to say hello
		
	// Supply voltage to the photo-transistor:
	pinMode( lightSensorSupplyPin, OUTPUT ); // genie.begin() does not set pin modes on any TRS, BNC or JST pins, so we have to do that manually
	digitalWrite( lightSensorSupplyPin, HIGH );

	//lightAdjustment.debug = "lightAdjustment"; soundAdjustment.debug = "soundAdjustment"; // TODO: for best performance, comment this line out - it causes Gauge event messages to be echoed as text to the main (USB) serial port
	
	// Set the `light` and `sound` instances' other triggering parameters here too, as desired.
}


void loop()
{
	unsigned long now = micros(); // most event detectors need a microsecond timestamp. We can save a few microseconds' overhead by reading the time once and passing it in, rather than making each of them read the clock.
	
	if( lightAdjustment.changed( now ) ) light.threshold = lightAdjustment / 10; // You can play with the denominators until the pots
	if( soundAdjustment.changed( now ) ) sound.threshold = soundAdjustment / 10; // feel like they have the right kind of dynamic range.
	// Turning a pot all the way to the right gives you the smallest analog readings -> lowest threshold -> greatest sensitivity.
	// Note that you may need to program a very different range of threshold values for a light-dependent resistor (LDR) than you do
	// for a photo-transistor.

	light.process( now ); // use the `TriggerDC` instance to process the DC light-sensor signal
	sound.process( now ); // use the `Trigger` instance to process the oscillatory audio signal
	// `sound.process(now)` can alternatively be broken down into `sound.read(now) && sound.write(now)`
	// The first call, `sound.read(now)`, returns `true` if a new sample was read with `analogRead()`,
	// or `false` if it was not yet time to read a new sample according to the specified
	// `sound.samplePeriodInMicros`. The second call, `sound.write(now)`, updates the trigger output
	// state according to the various other parameters (see the `Trigger_Parameters` and
	// `TriggerDC_Parameters` example sketches) and returns the current on/off pulse state.
	// (One slight difference is that `sound.process(now)` returns the on/off state regardless of
	// whether it was time to read a new sample or not.)
	
	
	// Deal with subject responses:
	switch( response.event() )
	{
		case Switch::PRESS:   digitalWrite( responseOutputPin, HIGH ); break;
		case Switch::RELEASE: digitalWrite( responseOutputPin, LOW  ); break;
		default: break;
	}
	// we could simply have sid `digitalWrite( responseOutputPin, response.isClosed() );` 
	// but the `event()` method has the advantage of debouncing (i.e. ignoring very short
	// release + press pairs)
	
	
	// Let's do the same again with `genie.pushButton`, which is a pre-configured `Switch`
	// instance that represents the SyncGenie's top-side push-button:
	switch( genie.pushButton.event() )
	{
		case Switch::PRESS:   digitalWrite( genie.D3_OUT, HIGH ); break;
		case Switch::RELEASE: digitalWrite( genie.D3_OUT, LOW  ); break;
		default: break;
	}	
}
