/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __Triggering_CPP__
#define   __Triggering_CPP__

#include <limits.h>

#include "Triggering.h"

Trigger::Trigger( int inputPin, int outputPin1, int outputPin2, double memoryInSeconds )
:
	samplePeriodInMicros( 100 ),
	inverted( false ),
	threshold( 15 ),
	recentMax( 0 ),
	minimumPulseDurationInMicros( 10000 ),
	pulseDurationIncrementInMicros( 10000 ),
	refractoryPeriodInMillis( 200 ),
	pulseStartCallback( NULL ),
	pulseFinishCallback( NULL ),

	mInputPin( inputPin ),
	mOutputPin1( outputPin1 ),
	mOutputPin2( outputPin2 ),

	mPreviousReading( 0 ),
	mSignal( 0 ),
	mPulseOn( false ),
	mInRefractoryPeriod( false ),
	mLastSampleTimeInMicros( 0 ),
	mTimeLastExceeded( 0 ),
	mLastPulseStartTimeInMicros( 0 ),
	mLastPulseFinishTimeInMicros( 0 ),
	mLastPulseDurationInMicros( 0 ),
	mPulsesStarted( 0 ),
	mPulsesFinished( 0 ),
	mSlidingWindow( NULL )
{
	this->rememberSamples( memoryInSeconds, 20 );
	this->begin();
}

Trigger::~Trigger() {}


void
Trigger::begin( void )
{
	if( mInputPin   >= 0 ) pinMode( mInputPin,   INPUT  );
	if( mOutputPin1 >= 0 ) pinMode( mOutputPin1, OUTPUT );
	if( mOutputPin2 >= 0 ) pinMode( mOutputPin2, OUTPUT );
}

void
Trigger::rememberSamples( double maxMemoryInSeconds, unsigned long nBins )
{
	// Assuming samplePeriodInMicros is 100, default settings give you bins
	// each spanning 500 samples, i.e. 50ms. Multiply that by 20 bins, and
	// you get a total memory span of 1 second in mSlidingWindow (although
	// one bin always gets emptied when all bins are full, so the memory span
	// actually fluctuates between 0.95 and 1 second).
	if( maxMemoryInSeconds < 0.0 ) maxMemoryInSeconds = 0.0;
	unsigned long samplesPerBin = ( unsigned long )( 0.5 + 1e6 * maxMemoryInSeconds / ( this->samplePeriodInMicros * nBins ) ); // microsPerTotal / microsPerSample / binsPerTotal
	if( !samplesPerBin ) { delete mSlidingWindow; mSlidingWindow = NULL; return; }
	if( mSlidingWindow ) mSlidingWindow->resize( samplesPerBin, nBins );
	else mSlidingWindow = new RecentMax< int >( samplesPerBin, nBins );
}

int
Trigger::getSignal( void )
{
	int currentReading = 0;
	if( mInputPin >= 0 ) currentReading = ::analogRead( mInputPin );
	int absDiff = abs( currentReading - mPreviousReading );
	mPreviousReading = currentReading;
	return absDiff;
}

bool
Trigger::read( unsigned long t )
{
	if( t - mLastSampleTimeInMicros < this->samplePeriodInMicros ) return false;
	mLastSampleTimeInMicros = t;
	mSignal = getSignal();
	return true;
}

bool
Trigger::force( unsigned long t, bool exceedThreshold )
{
	if( !this->pulseDurationIncrementInMicros ) // if the increment is 0, assume we only want transient pulses, and only when the forced signal rises
	{
		bool risen = !mPreviousReading && exceedThreshold;
		mPreviousReading = ( int )exceedThreshold;
		exceedThreshold = risen;
	}
	
	if( exceedThreshold ) // reset the sample timer - might as well deliver the pulse ASAP
	{
		mLastSampleTimeInMicros = t;
		mSignal = INT_MAX;
		return true;
	}
	// Otherwise, treat it like a normally scheduled sample	
	if( t - mLastSampleTimeInMicros < this->samplePeriodInMicros ) return false;
	mLastSampleTimeInMicros = t;
	mSignal = INT_MIN;
	return true;
}

bool
Trigger::write( unsigned long t )
{
	
	if( t != mLastSampleTimeInMicros ) return mPulseOn;
	unsigned long microsSinceLastPulseStart = mLastSampleTimeInMicros - mLastPulseStartTimeInMicros;
	
	// If we think we're still in the refractory period, but the refractory duration has
	// in fact elapsed, mark the refractory period as finished:
	if( mInRefractoryPeriod && microsSinceLastPulseStart >= this->refractoryPeriodInMillis * 1000 ) mInRefractoryPeriod = false;
	
	bool exceeded = this->inverted ?
		mSignal <= this->threshold :
		mSignal >= this->threshold; // OK, the >= operator doesn't quite match the "exceed" variable-naming semantics, but it made sense to change > to >= so that the edge-case threshold=0 means pulse-always-on
	if( exceeded ) mTimeLastExceeded = mLastSampleTimeInMicros;
	
	if( mPulseOn )
	{
		if( microsSinceLastPulseStart >= this->minimumPulseDurationInMicros && mLastSampleTimeInMicros - mTimeLastExceeded >= this->pulseDurationIncrementInMicros )
		{
			// The pulse always lasts 7--9 ms longer than the input both nominally (duration computed here) and objectively (when recorded with nidaq).
			// The extra duration depends on both the frequency content of the input, and the pulseDurationIncrementInMicros (but this cannot be made much shorter than 10ms)
			if( mOutputPin1 >= 0 ) digitalWrite( mOutputPin1, LOW );
			if( mOutputPin2 >= 0 ) digitalWrite( mOutputPin2, LOW );
			mPulseOn = false;
			mPulsesFinished++;
			mLastPulseFinishTimeInMicros = mLastSampleTimeInMicros;
			mLastPulseDurationInMicros = microsSinceLastPulseStart;
			if( this->pulseFinishCallback ) this->pulseFinishCallback( *this );
		}
	}
	else
	{
		if( !mInRefractoryPeriod && exceeded )
		{
			if( mOutputPin1 >= 0 ) digitalWrite( mOutputPin1, HIGH );
			if( mOutputPin2 >= 0 ) digitalWrite( mOutputPin2, HIGH );
			mPulseOn = true;
			mPulsesStarted++;
			mLastPulseStartTimeInMicros = mLastSampleTimeInMicros;
			if( this->refractoryPeriodInMillis > 0 ) mInRefractoryPeriod = true;
			if( this->pulseStartCallback ) this->pulseStartCallback( *this );
		}
	}
	
	if( mSlidingWindow ) this->recentMax = mSlidingWindow->process( mSignal );
	
	return mPulseOn;
}

TriggerDC::TriggerDC( int inputPin, int outputPin1, int outputPin2, double memoryInSeconds )
: Trigger( inputPin, outputPin1, outputPin2, memoryInSeconds ) {}

TriggerDC::~TriggerDC() {}

int
TriggerDC::getSignal( void )
{
	if( mInputPin < 0 ) return 0;
	return ::analogRead( mInputPin );
}

#endif // __Triggering_CPP__
